import 'dart:io';
import 'dart:async';
import 'dart:collection' show LinkedHashMap;
import 'dart:convert';
import 'business/encryption.dart';
import 'package:meta/meta.dart';
import 'package:steel_crypt/steel_crypt.dart' show CryptKey;

/// Generate a 32-byte, random Fortuna key. You can pass this to `init` if you
/// are using encryption.
String genKey () {
  return CryptKey().genFortuna();
}

/// Wrapper class for interacting with a database. These databases are organized
/// as a key-value store (similar to a Map).
class TailorDB {
  bool _beenChanged = false;
  bool _encryption;
  File _file;
  int _storeCacheInter;
  LinkedHashMap<String, dynamic> _cache;

  ///Returns whether this database is being encrypted.
  bool get encryption => _encryption;

  /// Initialize the database with a JSON file that it can write to for persistent
  /// storage. You can get this file using path_provider in Flutter.
  ///
  /// Pass true for `encryption` if you would like ISAAC encryption to be enabled
  /// this database. You must also pass a 32 byte keystring along with it. You can
  /// use `genKey` to generate this.
  ///
  /// Pass a refreshInterval to specify how often automatic pushes are done. The
  /// automatic pushes will happen every `refreshInterval` milliseconds.
  TailorDB({@required File file, bool encryption = false, String passkey = '', int refreshInterval = 250}) {
    _encryption = encryption;
    _storeCacheInter = refreshInterval;
    _init(file, encryption, passkey);
  }

  /// Adds `key`:`value` to end of database.
  ///
  /// `value` must be a `String`, `List<String>`, or `Map<String, String>`, or
  /// an error will be thrown.
  void add({@required String key, @required dynamic value}) {
    if (value is! String &&
        value is! List<String> &&
        value is! Map<String, String>) {
      throw ArgumentError('Invalid object type: ${value.runtimeType}. Only '
          'String, List<String> and Map<String, String> are supported.');
    }
    _append(key, value);
  }

  /// Returns true if the current cache state has been loaded into persistent
  /// storage. The frequency of these loads can be changed via the constructor.
  /// The loads can be manually induced via `writeCache`.
  bool get isFlushed => !_beenChanged;

  /// Get the value associated with this key. This can also be done using the
  /// `[]` operator.
  dynamic get ({@required String key}) => _get(key);

  /// Get the value associated with this key. This can also be done using the
  /// `get` method.
  dynamic operator [] (String key) => _get(key);

  /// Removes given key and associated value from database.
  void remove({@required String key}) {
    _remove(key);
  }

  /// Manually load the cache state into the database. It is automatically loaded
  /// at the rate specified in the constructor.
  void writeCache () {
    if (_beenChanged) {
      _file.writeAsStringSync(jsonEncode(_cache));
      _beenChanged = false;
    }
  }

  /// Returns whether this database contains given key.
  void containsKey ({@required String key}) => _containsKey(key);

  /// Returns whether this database contains given value.
  void containsValue ({@required dynamic value}) {
    if (value is! String || value is! List<String> || value is! Map<String, String>) {
      throw ArgumentError('Invalid object type: ${value.runtimeType}. Only '
          'String, List<String> and Map<String, String> are supported.');
    }
    _containsValue(value);
  }

  /// Returns whether this database is empty.
  bool get isEmpty => _cache.isEmpty;

  /// Returns the values of this database.
  Iterable get values => _values;

  /// Returns the keys of this database.
  Iterable get keys => _keys;

  /// Clear everything off this database.
  void clear() => _clear();

  /// Asynchronously adds `key`:`value` to end of database.
  ///
  /// `value` must be a `String`, `List<String>`, or `Map<String, String>`, or
  /// an error will be thrown.
  Future<void> addAsync({@required String key, @required dynamic value}) async {
    if (value is! String &&
        value is! List<String> &&
        value is! Map<String, String>) {
      throw ArgumentError('Invalid object type: ${value.runtimeType}. Only '
          'String, List<String> and Map<String, String> are supported.');
    }
    _append(key, value);
  }

  /// Asynchronously get the value associated with this key. This can also be done using the
  /// `[]` operator.
  Future<dynamic> getAsync ({@required String key}) async => _get(key);

  /// Asynchronously removes given key and associated value from database.
  Future<void> removeAsync ({@required String key}) async {
    _remove(key);
  }

  /// Asynchronously returns whether this database contains given key.
  Future<void> containsKeyAsync ({@required String key}) async => _containsKey(key);

  /// Asynchronously returns whether this database contains given value.
  Future<void> containsValueAsync ({@required dynamic value}) async {
    if (value is! String || value is! List<String> || value is! Map<String, String>) {
      throw ArgumentError('Invalid object type: ${value.runtimeType}. Only '
          'String, List<String> and Map<String, String> are supported.');
    }
    _containsValue(value);
  }

  /// Asynchronously returns the values of this database.
  Future<Iterable> get valuesAsync async => _values;

  /// Asynchronously returns the keys of this database.
  Future<Iterable> get keysAsync async => _keys;


  // internal methods
  void _init(File jsonFile, bool encryption, [String passkey]) async {
    _file = jsonFile;
    if (encryption) {
      _encryption = encryption;
      key = passkey;
    }
    try {
      _cache =
      LinkedHashMap<String, dynamic>.of(jsonDecode(_file.readAsStringSync()));
    } catch (on) {
      _cache = LinkedHashMap<String, dynamic>.of({});
    }
    _file.openSync(mode: FileMode.write);
    Timer.periodic(Duration(milliseconds:  _storeCacheInter), _store);
  }

  void _store (Timer timer) {
    writeCache();
  }

  void _append(String key, dynamic value) {
    if (_encryption == true) {
      if (value is String) {
        _cache.addAll({encrypt(key): encrypt(value)});
      } else if (value is List<String>) {
        _cache.addAll({encrypt(key): encryptList(value)});
      } else {
        _cache.addAll({encrypt(key): encryptMap(value)});
      }
    } else {
      _cache.addAll({key: value});
    }
    _beenChanged = true;
  }

  dynamic _get(String key) {
    if (_encryption) {
      var eval = _cache[key];
      switch (eval.runtimeType) {
        case String:
          {
            return decrypt(eval);
          }
          break;
        case Map:
          {
            return decryptMap(eval);
          }
          break;
        default:
          {
            return decryptList(eval);
          }
          break;
      }
    }
    return _cache[key];
  }

  void _remove(String key) {
    if (_encryption) {
      _cache.remove(encrypt(key));
    } else {
      _cache.remove(key);
    }
    _beenChanged = true;
  }

  Iterable get _keys {
    if (_encryption) {
      return decryptList(_cache.keys.toList());
    }
    return _cache.keys;
  }

  Iterable get _values {
    if (_encryption) {
      var vals = _cache.values.toList();
      return List.generate(vals.length, (int i) {
        dynamic value = vals[i];
        if (value is String) {
          return decrypt(value);
        } else if (value is List) {
          return decryptList(value);
        } else {
          return decryptMap(value);
        }
      });
    }
    return _cache.values;
  }

  bool _containsKey(String key) {
    if (_encryption) {
      return _cache.containsKey(encrypt(key));
    }
    return _cache.containsKey(key);
  }

  bool _containsValue(dynamic value) {
    if (_encryption == true) {
      if (value is String) {
        return _cache.containsValue(decrypt(value));
      } else if (value is List) {
        return _cache.containsValue(decryptList(value));
      } else {
        return _cache.containsValue(decryptMap(value));
      }
    }
    return _cache.containsValue(value);
  }

  void _clear() {
    _cache = LinkedHashMap<String, dynamic>.of({});
  }

}
